module Chu2 where

open import Data.List using (List; [_])
open import Foreign.Haskell using (Unit)
open import IO using (IO; run; lift; _>>=_; return; _>>_)
open import Data.Unit using (⊤)
import IO.Primitive as Prim
open import Data.String using (String)
open import Coinduction using (♯_)

data Tuple (A : Set)(B : Set) : Set where
    _,_ : A → B → Tuple A B

{-# IMPORT Chu2.FFI #-}
{-# COMPILED_DATA Tuple (,) (,) #-}

Field = String
Header = Tuple Field Field
Headers = List Header

data RequestMethod : Set where
  OPTIONS   : RequestMethod
  GET       : RequestMethod
  HEAD      : RequestMethod
  POST      : RequestMethod
  PUT       : RequestMethod
  DELETE    : RequestMethod
  TRACE     : RequestMethod
  CONNECT   : RequestMethod
  

{-# COMPILED_DATA RequestMethod Chu2.FFI.RequestMethod
    Chu2.FFI.OPTIONS
    Chu2.FFI.GET
    Chu2.FFI.HEAD
    Chu2.FFI.POST
    Chu2.FFI.PUT
    Chu2.FFI.DELETE
    Chu2.FFI.TRACE
    Chu2.FFI.CONNECT
#-}

data Chu2UrlScheme : Set where
  HTTP HTTPS : Chu2UrlScheme
  
{-# COMPILED_DATA Chu2UrlScheme Chu2.FFI.Chu2UrlScheme
    Chu2.FFI.HTTP
    Chu2.FFI.HTTPS
#-}

ScriptName    =  Field
PathInfo      =  Field
QueryField    =  Field
ServerName    =  Field
ServerPort    =  Field
HttpHeaders   =  Headers
Chu2Input     =  Field
Chu2Headers   =  Headers

private
  data EnvData : Set where
    envData : 
        RequestMethod  
      → ScriptName     
      → PathInfo       
      → QueryField    
      → ServerName     
      → ServerPort     
      → HttpHeaders    
      → Chu2UrlScheme  
      → Chu2Input      
      → Chu2Headers   
      → EnvData

  {-# COMPILED_DATA EnvData Chu2.FFI.Env Chu2.FFI.Env #-}

record Env : Set where
  constructor env
  field
    requestMethod  : RequestMethod  
    scriptName     : ScriptName     
    pathInfo       : PathInfo       
    queryField     : QueryField    
    serverName     : ServerName     
    serverPort     : ServerPort     
    httpHeaders    : HttpHeaders    
    chu2UrlScheme  : Chu2UrlScheme  
    chu2Input      : Chu2Input      
    chu2Headers    : Chu2Headers   
    

private
  envDataToEnv : EnvData → Env
  envDataToEnv
    ( envData
        requestMethod  
        scriptName     
        pathInfo       
        queryField    
        serverName     
        serverPort     
        httpHeaders    
        chu2UrlScheme  
        chu2Input      
        chu2Headers    
    ) =
  
    env
      requestMethod  
      scriptName     
      pathInfo       
      queryField    
      serverName     
      serverPort     
      httpHeaders    
      chu2UrlScheme  
      chu2Input      
      chu2Headers    

  
  envToEnvData : Env → EnvData
  envToEnvData
    ( env
        requestMethod  
        scriptName     
        pathInfo       
        queryField    
        serverName     
        serverPort     
        httpHeaders    
        chu2UrlScheme  
        chu2Input      
        chu2Headers    
    ) =
    envData
      requestMethod  
      scriptName     
      pathInfo       
      queryField    
      serverName     
      serverPort     
      httpHeaders    
      chu2UrlScheme  
      chu2Input      
      chu2Headers  

data Status : Set where
  OK                        : Status
  Created                   : Status
  Accepted                  : Status
  NoContent                 : Status
  MultipleChoices           : Status
  MovedPermanently          : Status
  SeeOther                  : Status
  NotModified               : Status
  MovedTemporarily          : Status
  BadRequest                : Status
  Unauthorized              : Status
  Forbidden                 : Status
  NotFound                  : Status
  MethodNotAllowed          : Status
  NotAcceptable             : Status
  Conflict                  : Status
  Gone                      : Status
  PreconditionFailed        : Status
  RequestEntityTooLarge     : Status
  RequestURItooLong         : Status
  UnsupportedMediaType      : Status
  NotImplemented            : Status
  ServiceUnavailable        : Status

{-# COMPILED_DATA Status Chu2.FFI.Status
  Chu2.FFI.OK
  Chu2.FFI.Created
  Chu2.FFI.Accepted
  Chu2.FFI.NoContent
  Chu2.FFI.MultipleChoices
  Chu2.FFI.MovedPermanently
  Chu2.FFI.SeeOther
  Chu2.FFI.NotModified
  Chu2.FFI.MovedTemporarily
  Chu2.FFI.BadRequest
  Chu2.FFI.Unauthorized
  Chu2.FFI.Forbidden
  Chu2.FFI.NotFound
  Chu2.FFI.MethodNotAllowed
  Chu2.FFI.NotAcceptable
  Chu2.FFI.Conflict
  Chu2.FFI.Gone
  Chu2.FFI.PreconditionFailed
  Chu2.FFI.RequestEntityTooLarge
  Chu2.FFI.RequestURItooLong
  Chu2.FFI.UnsupportedMediaType
  Chu2.FFI.NotImplemented
  Chu2.FFI.ServiceUnavailable
#-}

Body = Field

private
  data ResponseData : Set where
    responseData : 
        Status
      → Headers
      → Body
      → ResponseData

  {-# COMPILED_DATA ResponseData Chu2.FFI.Response Chu2.FFI.Response #-}


record Response : Set where
  constructor response
  field
    status  : Status
    headers : Headers
    body    : Body

private
  responseToResponseData : Response → ResponseData
  responseToResponseData 
    ( response
        status
        headers
        body
    ) =
  
    responseData
      status
      headers
      body
    

  responseDataToResponse : ResponseData → Response
  responseDataToResponse 
    ( responseData
        status
        headers
        body
    ) =
  
    response
      status
      headers
      body 
    
defaultResponse : Response
defaultResponse = 
  record
    {
      status = OK
    ; headers = [("Content-Type", "text/plain; charset=utf-8")]
    ; body = "Chu2!"
    }

Application = Env → IO Response
Middleware = Application → Application

private
  RawApplication = EnvData → Prim.IO ResponseData

RawMiddleware = RawApplication → RawApplication

private
  open import Function using (_$_)
 
  applicationToRawApplication : Application → RawApplication
  applicationToRawApplication app = 
    λ aEnvData → run $
      ♯ app (envDataToEnv aEnvData) >>= λ aResponse →
      ♯ return (responseToResponseData aResponse)

  rawApplicationToApplication : RawApplication → Application
  rawApplicationToApplication rawApp = 
    λ aEnv → ♯ lift (rawApp (envToEnvData aEnv)) >>= λ aRawResponse → 
             ♯ return (responseDataToResponse aRawResponse)


Handler = Application → IO ⊤
RawHandler = RawApplication → Prim.IO Unit

open import Function using (_∘_; _$_; id)


rawHandlerToHandler : RawHandler → Handler
rawHandlerToHandler rawHandler app = 
  ♯ lift (rawHandler (applicationToRawApplication app)) >> 
  ♯ return _
    



rawMiddlewareToMiddleware : RawMiddleware → Middleware
rawMiddlewareToMiddleware r app = rawApplicationToApplication $ r $ applicationToRawApplication app